@@ -2,12 +2,16 @@
 extern crate lazy_static;
 extern crate mime_guess;
 extern crate lazysort;
+extern crate brotli2;
 extern crate base64;
+extern crate flate2;
+extern crate bzip2;
 #[macro_use]
 extern crate clap;
 extern crate iron;
 extern crate time;
 extern crate url;
+extern crate md6;
 
 mod error;
 mod options;
